#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include "atomic_helpers.h"

#define DEBUG_ECALL 0x4442434E

typedef struct {
    long error;
    long value;
} sbiret;

void trap_handler();
void trap_handler_wrapper() __attribute__((naked));

void user_entry();

unsigned long strlen(const char* str) {
    unsigned long len = 0;
    while(str[len] != '\0') {
        len++;
    }

    return len;
}

int main() {
    csrw(stvec, trap_handler_wrapper);
    csrw(sepc, user_entry);

    unsigned long sstatus = csrr(sstatus);
    sstatus |= (1 << 5);
    csrw(sstatus, sstatus);

    // _start in entry.s will do the sret
}

void user_entry(void) {
    while(1) {
        __asm__ volatile("ecall");
        __asm__ volatile("unimp");
    }
}

__attribute__((always_inline)) 
inline sbiret sbi_ecall6(unsigned long number, unsigned long arg0, unsigned long arg1, unsigned long arg2, unsigned long arg3, unsigned long arg4, unsigned long arg5) {
    sbiret ret;

    __asm__ volatile("mv a7, %[number]\n"
                     "li a6, 0\n"
                     "mv a0, %[arg0]\n"
                     "mv a1, %[arg1]\n"
                     "mv a2, %[arg2]\n"
                     "mv a3, %[arg3]\n"
                     "mv a4, %[arg4]\n"
                     "mv a5, %[arg5]\n"
                     "ecall\n"
                     "mv %[err], a0\n"
                     "mv %[val], a1\n"
        : [err] "=r"(ret.error), [val] "=r"(ret.value)
        : [number] "r"(number), [arg0] "r"(arg0), [arg1] "r"(arg1),
[arg2] "r"(arg2),[arg3] "r"(arg3), [arg4] "r"(arg4),[arg5] "r"(arg5)  
        : "a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7");

    return ret;
}

__attribute__((always_inline)) 
inline sbiret sbi_ecall5(unsigned long number, unsigned long arg0, unsigned long arg1, unsigned long arg2, unsigned long arg3, unsigned long arg4) {
    return sbi_ecall6(number, arg0, arg1, arg2, arg3, arg4, 0);
}

__attribute__((always_inline)) 
inline sbiret sbi_ecall4(unsigned long number, unsigned long arg0, unsigned long arg1, unsigned long arg2, unsigned long arg3) {
    return sbi_ecall5(number, arg0, arg1, arg2, arg3, 0);
}

__attribute__((always_inline)) 
inline sbiret sbi_ecall3(unsigned long number, unsigned long arg0, unsigned long arg1, unsigned long arg2) {
    return sbi_ecall4(number, arg0, arg1, arg2, 0);
}

__attribute__((always_inline)) 
inline sbiret sbi_ecall2(unsigned long number, unsigned long arg0, unsigned long arg1) {
    return sbi_ecall3(number, arg0, arg1, 0);
}

__attribute__((always_inline))
inline sbiret sbi_ecall1(unsigned long number, unsigned long arg0) {
    return sbi_ecall2(number, arg0, 0);
}

__attribute__((always_inline)) 
inline sbiret sbi_ecall0(unsigned long number) {
    return sbi_ecall1(number, 0);
}

sbiret console_debug_write(const char* str, unsigned long len) {
    return sbi_ecall2(DEBUG_ECALL, len, (unsigned long)str);
}

int puts(const char* str) {
    sbiret ret;
    if((ret = console_debug_write(str, strlen(str))).error != 0) {
        return -1;
    }
    
    if((ret = console_debug_write("\n", 1)).error != 0) {
        return -1;
    }
    return 0;
}

void trap_handler_wrapper() {
    __asm__ volatile("call trap_handler\n"
                     "sret");
}

void trap_handler(void) {
    unsigned long sepc = csrr(sepc);
    unsigned long scause = csrr(scause);
    unsigned long sstatus = csrr(sstatus);

    if(scause == 8) {
        puts("hello world");

        sepc += 4;
        csrw(sepc, sepc);
        goto exit;
    }

    if(((sstatus >> 8) & 1) == 0) {
        sepc += 4;
        puts("exception occured in usermode, skipping past");
        csrw(sepc, sepc);
        goto exit;
    } else {
        puts("exception occured in supervisor mode, panic");

        while(1) {
            __asm__ volatile("wfi");
        }
    }

exit:
    return;
}
